#!/usr/bin/python3

import os
import gettext

def strip_split_and_recombine(comma_separated):
    word_list = comma_separated.split(",")
    out = ""
    for item in word_list:
        out += item.strip()
        out+=";"

    return out

def generate(domain, path, filename, prefix, name, comment, suffix, genericName=None, keywords=None, append=False):
    if append:
        desktopFile = open(filename, "a")
    else:
        desktopFile = open(filename, "w")

    desktopFile.writelines(prefix)

    desktopFile.writelines("Name=%s\n" % name)
    for directory in sorted(os.listdir(path)):
        mo_file = os.path.join(path, directory, "LC_MESSAGES", "%s.mo" % domain)
        if os.path.exists(mo_file):
            try:
                language = gettext.translation(domain, path, languages=[directory])
                L_ = language.gettext
                if (L_(name) != name):
                    desktopFile.writelines("Name[%s]=%s\n" % (directory, L_(name)))
            except:
                pass

    if comment is not None:
        desktopFile.writelines("Comment=%s\n" % comment)
        for directory in sorted(os.listdir(path)):
            mo_file = os.path.join(path, directory, "LC_MESSAGES", "%s.mo" % domain)
            if os.path.exists(mo_file):
                try:
                    language = gettext.translation(domain, path, languages=[directory])
                    L_ = language.gettext
                    if (L_(comment) != comment):
                        desktopFile.writelines("Comment[%s]=%s\n" % (directory, L_(comment)))
                except:
                    pass

    if keywords is not None:
        formatted = strip_split_and_recombine(keywords)
        desktopFile.writelines("Keywords=%s\n" % formatted)
        for directory in sorted(os.listdir(path)):
            mo_file = os.path.join(path, directory, "LC_MESSAGES", "%s.mo" % domain)
            if os.path.exists(mo_file):
                try:
                    language = gettext.translation(domain, path, languages=[directory])
                    L_ = language.gettext
                    if (L_(keywords) != keywords):
                        translated = strip_split_and_recombine(L_(keywords))
                        desktopFile.writelines("Keywords[%s]=%s\n" % (directory, translated))
                except:
                    pass

    if genericName is not None:
        desktopFile.writelines("GenericName=%s\n" % genericName)
        for directory in sorted(os.listdir(path)):
            mo_file = os.path.join(path, directory, "LC_MESSAGES", "%s.mo" % domain)
            if os.path.exists(mo_file):
                try:
                    language = gettext.translation(domain, path, languages=[directory])
                    L_ = language.gettext
                    if (L_(genericName) != genericName):
                        desktopFile.writelines("GenericName[%s]=%s\n" % (directory, L_(genericName)))
                except:
                    pass

    desktopFile.writelines(suffix)
    desktopFile.close()

def generate_polkit_policy(domain, path, filename, prefix, message, suffix, append=False):
    if append:
        policyFile = open(filename, "a")
    else:
        policyFile = open(filename, "w")

    policyFile.writelines(prefix)

    policyFile.writelines("<message>%s</message>\n" % message)
    for directory in sorted(os.listdir(path)):
        mo_file = os.path.join(path, directory, "LC_MESSAGES", "%s.mo" % domain)
        if os.path.exists(mo_file):
            try:
                language = gettext.translation(domain, path, languages=[directory])
                L_ = language.gettext
                if (L_(message) != message):
                    policyFile.writelines("<message xml:lang=\"%s\">%s</message>\n" % (directory, L_(message)))
            except:
                pass

    policyFile.writelines(suffix)
    policyFile.close()



import time
import os
from pathlib import Path
import json
import threading

import gi
gi.require_version('AppStream', '1.0')
from gi.repository import GLib, GObject

from . import _apt
from . import _flatpak
from ._flatpak import FlatpakRemoteInfo
from .pkgInfo import FlatpakPkgInfo, AptPkgInfo
from .misc import print_timing

SYS_CACHE_PATH = "/var/cache/mintinstall/pkginfo.json"
USER_CACHE_PATH = os.path.join(GLib.get_user_cache_dir(), "mintinstall", "pkginfo.json")

MAX_AGE = 7 * (60 * 60 * 24) # days

class CacheLoadingException(Exception):
    '''Thrown when there was an issue loading the pickled package set'''

class JsonObject(object):
    def __init__(self, pkginfo_cache, section_lists, flatpak_remote_infos):
        super(JsonObject, self).__init__()

        self.pkginfo_cache = pkginfo_cache
        self.section_lists = section_lists
        self.flatpak_remote_infos = flatpak_remote_infos

    @classmethod
    def from_json(cls, json_data: dict):
        pkgcache_dict = {}
        for key in json_data["pkginfo_cache"].keys():
            pkginfo_data = json_data["pkginfo_cache"][key]

            if pkginfo_data["pkg_hash"].startswith("a"):
                pkgcache_dict[key] = AptPkgInfo.from_json(pkginfo_data)
            else:
                pkgcache_dict[key] = FlatpakPkgInfo.from_json(pkginfo_data)

        remotes_dict = {}
        for key in json_data["flatpak_remote_infos"].keys():
            remote_data = json_data["flatpak_remote_infos"][key]
            remotes_dict[key] = FlatpakRemoteInfo.from_json(remote_data)

        return cls(pkgcache_dict,
                   json_data["section_lists"],
                   remotes_dict)

    def to_json(self):
        return self.__dict__

class PkgCache(object):
    STATUS_EMPTY = 0
    STATUS_OK = 1

    @print_timing
    def __init__(self, have_flatpak=False):
        super(PkgCache, self).__init__()

        self.status = self.STATUS_EMPTY
        self.have_flatpak = have_flatpak

        self._items = {}
        self._item_lock = threading.Lock()

        try:
            cache, sections, flatpak_remote_infos = self._load_cache()
        except CacheLoadingException:
            cache = {}
            sections = {}
            flatpak_remote_infos = {}

        if len(cache) > 0:
            self.status = self.STATUS_OK
        else:
            self.status = self.STATUS_EMPTY

        self._items = cache
        self.sections = sections
        self.flatpak_remote_infos = flatpak_remote_infos

    def keys(self):
        with self._item_lock:
            return self._items.keys()

    def values(self):
        with self._item_lock:
            return self._items.values()

    def __getitem__(self, key):
        with self._item_lock:
            return self._items[key]

    def __setitem__(self, key, value):
        with self._item_lock:
            self._items[key] = value

    def __delitem__(self, key):
        with self._item_lock:
            del self._items[key]

    def __contains__(self, pkg_hash):
        with self._item_lock:
            return pkg_hash in self._items

    def __len__(self):
        with self._item_lock:
            return len(self._items)

    def __iter__(self):
        with self._item_lock:
            for pkg_hash in self._items:
                yield self[pkg_hash]
            return

    def _generate_cache(self):
        cache = {}
        sections = {}
        flatpak_remote_infos = {}

        if self.have_flatpak:
            cache, flatpak_remote_infos = _flatpak.process_full_flatpak_installation(cache)

        cache, sections = _apt.process_full_apt_cache(cache)

        return cache, sections, flatpak_remote_infos

    def _get_best_load_path(self):
        try:
            sys_mtime = os.path.getmtime(SYS_CACHE_PATH)

            if ((time.time() - MAX_AGE) > sys_mtime) or not os.access(SYS_CACHE_PATH, os.R_OK):
                print("Installer: System pkgcache too old or not accessible, skipping")
                sys_mtime = 0
        except OSError:
            sys_mtime = 0

        try:
            user_mtime = os.path.getmtime(USER_CACHE_PATH)

            if (time.time() - MAX_AGE) > user_mtime:
                print("Installer: User pkgcache too old, skipping")
                user_mtime = 0
        except OSError:
            user_mtime = 0

        # If neither exist, return None, and a new cache will be generated
        if sys_mtime == 0 and user_mtime == 0:
            return None

        most_recent = None

        # Select the most recent
        if sys_mtime > user_mtime:
            most_recent = SYS_CACHE_PATH
            print("Installer: System pkgcache is most recent, using it.")
        else:
            most_recent = USER_CACHE_PATH
            print("Installer: User pkgcache is most recent, using it.")

        return Path(most_recent)

    @print_timing
    def _load_cache(self):
        """
        The cache pickle file can be in either a system or user location,
        depending on how the cache was generated.  If it exists in both places, take the
        most recent one.  If it's more than MAX_AGE, generate a new one anyhow.
        """

        cache = None
        sections = None
        flatpak_remote_infos = None

        path = self._get_best_load_path()

        if path is None:
            raise CacheLoadingException
        try:
            with path.open(mode='r', encoding="utf8") as f:
                json_obj = JsonObject.from_json(json.load(f))
                cache = json_obj.pkginfo_cache
                sections = json_obj.section_lists
                flatpak_remote_infos = json_obj.flatpak_remote_infos
        except Exception as e:
            print("Installer: Error loading pkginfo cache:", e)
            cache = None

        if cache == None:
            raise CacheLoadingException

        return cache, sections, flatpak_remote_infos

    def _get_best_save_path(self):
        best_path = None

        # Prefer the system location, as all users can access it
        try:
            path = Path(SYS_CACHE_PATH)
            path.parent.mkdir(parents=True, exist_ok=True)
            path.touch(exist_ok=True)
        except PermissionError:
            try:
                path = Path(USER_CACHE_PATH)
                path.parent.mkdir(parents=True, exist_ok=True)
            except Exception:
                path = None
        finally:
            best_path = path

        return best_path

    def _save_cache(self, to_be_json):
        path = self._get_best_save_path()

        FlatpakPkgInfo.__module__ = "installer.pkgInfo"
        AptPkgInfo.__module__ = "installer.pkgInfo"
        FlatpakRemoteInfo.__module__ = "installer._flatpak"

        try:
            with path.open(mode='w', encoding="utf8") as f:
                json.dump(to_be_json, f, default=lambda o: o.to_json(), indent=4)
        except Exception as e:
            print("Installer: Could not save cache:", str(e))

    def _new_cache_common(self):
        print("Installer: Generating new pkgcache")
        cache, sections, flatpak_remote_infos = self._generate_cache()

        if len(cache) > 0:
            self._save_cache(JsonObject(cache, sections, flatpak_remote_infos))

        with self._item_lock:
            self._items = cache
            self.sections = sections
            self.flatpak_remote_infos = flatpak_remote_infos

        if len(cache) == 0:
            self.status = self.STATUS_EMPTY
        else:
            self.status = self.STATUS_OK

    def _generate_cache_thread(self, callback=None):
        self._new_cache_common()

        if callback != None:
            GObject.idle_add(callback)

    def get_subset_of_type(self, pkg_type):
        with self._item_lock:
            return {k: v for k, v in self._items.items() if k.startswith(pkg_type)}

    def force_new_cache_async(self, idle_callback=None):
        thread = threading.Thread(target=self._generate_cache_thread,
                                  kwargs={"callback" : idle_callback})
        thread.start()

    def force_new_cache(self):
        self._new_cache_common()

    def find_pkginfo(self, string, pkg_type=None):
        if pkg_type in (None, "a"):
            pkginfo = _apt.find_pkginfo(self, string)

            if pkginfo != None:
                return pkginfo

        if self.have_flatpak:
            if pkg_type in (None, "f"):
                pkginfo = _flatpak.find_pkginfo(self, string)

                if pkginfo != None:
                    return pkginfo

        return None

    def _get_manually_installed_debs(self):
        """
        Generate list of manually installed Debian package.
        Requires a package list provided by the installer.
            Currently knows only Ubiquity's /var/log/installer/initial-status.gz
        """
        installer_log = "/var/log/installer/initial-status.gz"
        if not os.path.isfile(installer_log):
            return None
        import gzip
        try:
            installer_log = gzip.open(installer_log, "r").read().decode('utf-8').splitlines()
        except Exception as e:
            # There are a number of different exceptions here, but there's only one response
            print("Could not get initial installed packages list (check /var/log/installer/initial-status.gz): %s" % str(e))
            return None
        initial_status = [x[9:] for x in installer_log if x.startswith("Package: ")]
        if not initial_status:
            return None
        from . import _apt
        pkgcache = [x[4:] for x in self.get_subset_of_type("a")]
        current_status = ["apt:%s" % pkg for pkg in _apt.get_apt_cache() if
            (pkg.installed and
            not pkg.is_auto_installed and
            pkg.shortname not in initial_status and
            pkg.shortname in pkgcache)]
        return current_status

    def get_manually_installed_packages(self):
        """ Get list of all manually installed packages (apt and flatpak) """
        installed_packages = None
        installed_packages_apt = self._get_manually_installed_debs()
        if installed_packages_apt:
            installed_packages = installed_packages_apt
            installed_packages += [x for x in self.get_subset_of_type("f")]
        return installed_packages




#!/usr/bin/python3

import gettext
import os
import subprocess
import sys

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

import mintcommon.aptdaemon

# i18n
gettext.install("mint-common", "/usr/share/linuxmint/locale")

class MintRemoveWindow:

    def __init__(self, desktopFile):
        self.desktopFile = desktopFile
        process = subprocess.run(["dpkg", "-S", self.desktopFile], stdout=subprocess.PIPE)
        output = process.stdout.decode("utf-8")
        package = output[:output.find(":")].split(",")[0]
        if process.returncode != 0:
            if not self.try_remove_flatpak(desktopFile):
                warnDlg = Gtk.MessageDialog(None, 0, Gtk.MessageType.WARNING, Gtk.ButtonsType.YES_NO, _("This menu item is not associated to any package. Do you want to remove it from the menu anyway?"))
                warnDlg.set_keep_above(True)

                warnDlg.get_widget_for_response(Gtk.ResponseType.YES).grab_focus()
                warnDlg.vbox.set_spacing(10)
                response = warnDlg.run()
                if response == Gtk.ResponseType.YES:
                    print ("removing '%s'" % self.desktopFile)
                    subprocess.run(["rm", "-f", self.desktopFile])
                    subprocess.run(["rm", "-f", "%s.desktop" % self.desktopFile])
                warnDlg.destroy()

            sys.exit(0)

        warnDlg = Gtk.MessageDialog(None, 0, Gtk.MessageType.WARNING, Gtk.ButtonsType.OK_CANCEL, _("The following packages will be removed:"))
        warnDlg.set_keep_above(True)

        warnDlg.get_widget_for_response(Gtk.ResponseType.OK).grab_focus()
        warnDlg.vbox.set_spacing(10)

        treeview = Gtk.TreeView()
        column1 = Gtk.TreeViewColumn(_("Packages to be removed"))
        renderer = Gtk.CellRendererText()
        column1.pack_start(renderer, False)
        column1.add_attribute(renderer, "text", 0)
        treeview.append_column(column1)

        packages = []
        model = Gtk.ListStore(str)
        dependenciesString = subprocess.getoutput("apt-get -s -q remove " + package + " | grep Remv")
        dependencies = dependenciesString.split("\n")
        for dependency in dependencies:
            dependency = dependency.replace("Remv ", "")
            model.append([dependency])
            packages.append(dependency.split()[0])
        treeview.set_model(model)
        treeview.show()

        scrolledwindow = Gtk.ScrolledWindow()
        scrolledwindow.set_shadow_type(Gtk.ShadowType.ETCHED_OUT)
        scrolledwindow.set_size_request(150, 150)
        scrolledwindow.add(treeview)
        scrolledwindow.show()

        warnDlg.get_content_area().add(scrolledwindow)

        self.apt = mintcommon.aptdaemon.APT(warnDlg)

        response = warnDlg.run()
        if response == Gtk.ResponseType.OK:
            self.apt.set_finished_callback(self.on_finished)
            self.apt.remove_packages(packages)
        elif response == Gtk.ResponseType.CANCEL:
            sys.exit(0)

        warnDlg.destroy()

    def try_remove_flatpak(self, desktopFile):
        if not "flatpak" in desktopFile:
            return False

        if not os.path.exists('/usr/bin/mintinstall-remove-app'):
            return False

        flatpak_remover = subprocess.Popen(['/usr/bin/mintinstall-remove-app', desktopFile])
        retcode = flatpak_remover.wait()

        return retcode == 0

    def on_finished(self, transaction=None, exit_state=None):
        sys.exit(0)

if __name__ == "__main__":

    # Exit if the given path does not exist
    if len(sys.argv) < 2 or not os.path.exists(sys.argv[1]):
        sys.exit(1)

    mainwin = MintRemoveWindow(sys.argv[1])
    Gtk.main()



#!/usr/bin/python3
"""
apt_changelog.py

Replacement for apt changelog, can be used both standalone and as a module.

In addition to apt changelog sources, it directly supports also Linux Mint
and Launchpad repositories and will try to retrieve a changelog for packages
from all other debian package sources as well. Most importantly, changelogs
for installed packages are retrieved locally to avoid unnecessary network
activity.

Unlike apt it does not support wildcards for the package name. You get to
check one specific changelog at a time only. I see no value in multi-lookups.

Copyright (c) 2018 gm10

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import fnmatch
import gzip
import lzma
import os
import sys
import tarfile
import tempfile

import requests
from apt.cache import Cache
from apt.debfile import DebPackage

import apt_pkg


class _Package():
    def __init__(self,name:str, fullname:str, architecture:str, version:str,
        source_name:str, source_version:str, uri:str, filename:str, origin:str,
        component:str, downloadable:bool, is_installed:bool, dependencies:list):
        self.name = name
        self.fullname = fullname
        self.architecture = architecture
        self.version = version
        self.source_name = source_name
        self.source_version = source_version
        self.source_version_raw = None
        self.uri = uri
        self.filename = filename
        self.origin = origin
        self.component = component
        self.downloadable = downloadable
        self.is_installed = is_installed
        self.installed_files = list()
        self.dependencies = dependencies

class _AptChangelog():

    def __init__(self, interactive:bool=False):
        self.interactive = interactive

        # constants
        # apt uses MB rather than MiB, so let's stay consistent
        self.MB = 1000 ** 2
        # downloads larger than this require confirmation or fail
        self.max_download_size_default = 1.5 * self.MB
        self.max_download_size = self.max_download_size_default
        max_download_size_msg_template = "\
To retrieve the full changelog, %s MB have to be downloaded.\n%s\
\n\
Proceed with the download?"
        self.max_download_size_msg_lc = max_download_size_msg_template % ("%.1f",
            "Otherwise we will try to retrieve just the last change.\n")
        self.max_download_size_msg = max_download_size_msg_template % ("%.1f","")
        self.max_download_size_msg_unknown = max_download_size_msg_template % ("an unknown amount of", "")

        self.apt_cache = None
        self.apt_cache_date = None
        self.candidate = None

        # get apt's configuration
        apt_pkg.init_config()
        if apt_pkg.config.exists("Acquire::Changelogs::URI::Origin"):
            self.apt_origins = apt_pkg.config.subtree("Acquire::Changelogs::URI::Origin")
        else:
            self.apt_origins = None
        if apt_pkg.config.exists("Dir::Cache::pkgcache"):
            self.apt_cache_path = apt_pkg.config.find_dir("Dir::Cache")
            self.pkgcache = apt_pkg.config.find_file("Dir::Cache::pkgcache")
        else:
            self.apt_cache = "invalid"
        if (self.apt_cache or
            not os.path.isdir(self.apt_cache_path) or
            not os.path.isfile(self.pkgcache)
            ):
            print("E: Invalid APT configuration found, try to run `apt update` first",
                file=sys.stderr)
            self.close(99)

    def get_cache_date(self):
        if os.path.isfile(self.pkgcache):
            return os.path.getmtime(self.pkgcache)
        return None

    def refresh_cache(self):
        cache_date = self.get_cache_date()

        if not self.apt_cache:
            self.apt_cache = Cache()
            self.apt_cache_date = cache_date
        elif cache_date != self.apt_cache_date:
            self.apt_cache.open(None)
            self.apt_cache_date = cache_date

    def drop_cache(self):
        if self.candidate:
            self.candidate = None
        self.apt_cache = None

    def get_changelog(self, pkg_name:str, no_local:bool=False):
        self.refresh_cache()
        self.candidate = self.parse_package_metadata(pkg_name)

        # parse the package's origin
        if not self.candidate.downloadable:
            origin = "local_package"
        elif self.candidate.origin == "linuxmint":
            origin = "linuxmint"
        elif self.candidate.origin.startswith("LP-PPA-"):
            origin = "LP-PPA"
        elif self.apt_origins and self.candidate.origin in self.apt_origins.list():
            origin = "APT"
        else:
            origin = "unsupported"

        # Check for changelog of installed package first
        has_local_changelog = False
        uri = None
        if not no_local and self.candidate.is_installed:
            if _DEBUG: print("Package is installed...")
            uri = self.get_changelog_from_filelist(
                self.candidate.installed_files, local=True)
            # Ubuntu kernel workarounds
            if self.candidate.origin == "Ubuntu":
                if self.candidate.source_name == "linux-signed":
                    uri = uri.replace("linux-image","linux-modules")
                if self.candidate.source_name == "linux-meta":
                    uri = None
            if uri and not os.path.isfile(uri):
                uri = None

        # Do nothing if local changelog exists
        if uri:
            has_local_changelog = True
        # all origins that APT supports
        elif origin == 'APT':
            uri = self.get_apt_changelog_uri(
                self.apt_origins.get(self.candidate.origin))
            r = self.check_url(uri)
            if not r:
                self.exit_on_fail(2)
        # Linux Mint repo
        elif origin == 'linuxmint':
            # Mint repos don't have .debian.tar.xz files, only full packages, so
            # check the package cache first
            base_uri, _ = os.path.split(self.candidate.uri)
            r, uri = self.get_changelog_uri(base_uri)
            if not r:
                # fall back to last change info for the source package
                # Mint's naming scheme seems to be using amd64 unless source
                # is i386 only, we always check amd64 first
                base_uri = "http://packages.linuxmint.com/dev/%s_%s_%s.changes"
                uri = base_uri % (self.candidate.source_name,
                    self.candidate.source_version, "amd64")
                r = self.check_url(uri, False)
                if not r:
                    uri = base_uri % (self.candidate.source_name,
                        self.candidate.source_version, "i386")
                    r = self.check_url(uri, False)
                    if not r:
                        self.exit_on_fail(3)

        # Launchpad PPA
        elif origin == 'LP-PPA':
            ppa_owner, ppa_name, _ = \
                self.candidate.uri.split("ppa.launchpad.net/")[1].split("/", 2)
            base_uri = "http://ppa.launchpad.net/%s/%s/ubuntu/pool/main/{self.source_prefix()}/%s" % (ppa_owner, ppa_name, self.candidate.source_name)
            r, uri = self.get_changelog_uri(base_uri)
            if not r:
                # fall back to last change info only
                uri = "https://launchpad.net/~%s/+archive/ubuntu/%s/+files/%s_%s_source.changes" % (ppa_owner, ppa_name, self.candidate.source_name, self.candidate.source_version)
                r = self.check_url(uri, False)
                if not r:
                    self.exit_on_fail(4)
        # Not supported origin
        elif origin == 'unsupported':
            if _DEBUG: print("Unsupported Package")
            base_uri, _ = os.path.split(self.candidate.uri)
            r, uri = self.get_changelog_uri(base_uri)
            if not r:
                self.exit_on_fail(5)
        # Locally installed package without local changelog or remote
        # source, hope it's cached and contains a changelog
        elif origin == 'local_package':
            uri = self.apt_cache_path + self.candidate.filename
            if not os.path.isfile(uri):
                self.exit_on_fail(6)

        # Changelog downloading, extracting and processing:
        changelog = ""
        # local changelog
        if has_local_changelog and not no_local:
            if _DEBUG: print("Using local changelog:",uri)
            try:
                filename = os.path.basename(uri)
                # determine file type by name/extension
                # as per debian policy 4.4 the encoding must be UTF-8
                # as per policy 12.7 the name must be changelog.Debian.gz or
                # changelog.gz (deprecated)
                if filename.lower().endswith('.gz'):
                    changelog = gzip.open(uri,'r').read().decode('utf-8')
                elif filename.lower().endswith('.xz'):
                    # just in case / future proofing
                    changelog = lzma.open(uri,'r').read().decode('utf-8')
                elif filename.lower() == 'changelog':
                    changelog = open(uri, 'r').read().encode().decode('utf-8')
                else:
                    raise ValueError('Unknown changelog format')
            except Exception as e:
                _generic_exception_handler(e)
                self.exit_on_fail(1)
        # APT-format changelog, download directly
        # - unfortunately this is slow since the servers support no compression
        elif origin == "APT":
            if _DEBUG: print("Downloading: %s (%.2f MB)" % (uri, r.length / self.MB))
            changelog = r.text
            r.close()
        # last change changelog, download directly
        elif uri.endswith('.changes'):
            if _DEBUG: print("Downloading: %s (%.2f MB)" % (uri, r.length / self.MB))
            changes = r.text.split("Changes:")[1].split("Checksums")[0].split("\n")
            r.close()
            for change in changes:
                change = change.strip()
                if change:
                    if change == ".":
                        change = ""
                    changelog += change + "\n"
        # compressed binary source, download and extract changelog
        else:
            source_is_cache = uri.startswith(self.apt_cache_path)
            if _DEBUG: print("Using cached package:" if source_is_cache else
                "Downloading: %s (%.2f MB)" % (uri, r.length / self.MB))
            try:
                if not source_is_cache:
                    # download stream to temporary file
                    tmpFile = tempfile.NamedTemporaryFile(prefix="apt-changelog-")
                    if self.interactive and r.length:
                        # download chunks with progress indicator
                        recv_length = 0
                        blocks = 60
                        for data in r.iter_content(chunk_size=16384):
                            recv_length += len(data)
                            tmpFile.write(data)
                            recv_pct = recv_length / r.length
                            recv_blocks = int(blocks * recv_pct)
                            print("\r[%(progress)s%(spacer)s] %(percentage).1f%%" %
                                {
                                    "progress": "=" * recv_blocks,
                                    "spacer":  " " * (blocks - recv_blocks),
                                    "percentage": recv_pct * 100
                                }, end="", flush=True)
                        # clear progress bar when done
                        print("\r" + " " * (blocks + 10), end="\r", flush=True)
                    else:
                        # no content-length or non-interactive, download in one go
                        # up to the configured max_download_size, ask only when
                        # exceeded
                        r.raw.decode_content = True
                        size = 0
                        size_exceeded = False
                        while True:
                            buf = r.raw.read(16*1024)
                            if not size_exceeded:
                                size += len(buf)
                                if size > self.max_download_size:
                                    if not self.user_confirm(self.max_download_size_msg_unknown):
                                        r.close()
                                        tmpFile.close()
                                        return ""
                                    else:
                                        size_exceeded = True
                            if not buf:
                                break
                            tmpFile.write(buf)
                    r.close()
                    tmpFile.seek(0)
                if uri.endswith(".deb"):
                    # process .deb file
                    if source_is_cache:
                        f = uri
                    else:
                        f = tmpFile.name
                        # We could copy the downloaded .deb files to the apt
                        # cache here but then we'd need to run the script elevated:
                        # shutil.copy(f, self.apt_cache_path + os.path.basename(uri))
                    deb = DebPackage(f)
                    changelog_file = self.get_changelog_from_filelist(deb.filelist)
                    if changelog_file:
                        changelog = deb.data_content(changelog_file)
                        if changelog.startswith('Automatically decompressed:'):
                            changelog = changelog[29:]
                    else:
                        raise ValueError('Malformed Debian package')
                elif uri.endswith(".diff.gz"):
                    # Ubuntu partner repo has .diff.gz files,
                    # we can extract a changelog from that
                    data = gzip.open(tmpFile.name, "r").read().decode('utf-8')
                    additions = data.split("+++")
                    for addition in additions:
                        lines = addition.split("\n")
                        if "/debian/changelog" in lines[0]:
                            for line in lines[2:]:
                                if line.startswith("+"):
                                    changelog += "%s\n" % line[1:]
                                else:
                                    break
                    if not changelog:
                        raise ValueError('No changelog in .diff.gz')
                else:
                    # process .tar.xz file
                    with tarfile.open(fileobj=tmpFile, mode="r:xz") as tar:
                        changelog_file = self.get_changelog_from_filelist(
                            [s.name for s in tar.getmembers() if s.type in (b"0", b"2")])
                        if changelog_file:
                            changelog = tar.extractfile(changelog_file).read().decode()
                        else:
                            raise ValueError('No changelog in source package')
            except Exception as e:
                _generic_exception_handler(e)
                self.exit_on_fail(520)
            if 'tmpFile' in vars():
                try:
                    tmpFile.close()
                except Exception as e:
                    _generic_exception_handler(e)

        # ALL DONE
        return changelog

    def parse_package_metadata(self, pkg_name:str):
        """ Creates the self.candidate object based on package name=version/release

        Wildcard matching is only used for version and release, and only the
        first match is processed.
        """
        # parse =version declaration
        if "=" in pkg_name:
            (pkg_name, pkg_version) = pkg_name.split("=", 1)
            pkg_release = None
        # parse /release declaration (only if no version specified)
        elif "/" in pkg_name:
            (pkg_name, pkg_release) = pkg_name.split("/", 1)
            pkg_version = None
        else:
            pkg_version = None
            pkg_release = None

        # check if pkg_name exists
        # unlike apt no pattern matching, a single exact match only
        if pkg_name in self.apt_cache:
            pkg = self.apt_cache[pkg_name]
        else:
            print("E: Unable to locate package %s" % pkg_name, file=sys.stderr)
            self.close(13)

        # get package data
        _candidate = None
        candidate = None
        if pkg_release or pkg_version:
            match_found = False
            for _pkg in pkg.versions:
                if pkg_version:
                    if fnmatch.fnmatch(_pkg.version, pkg_version):
                        match_found = True
                else:
                    for _origin in _pkg.origins:
                        if fnmatch.fnmatch(_origin.archive, pkg_release):
                            match_found = True
                if match_found:
                    _candidate = _pkg
                    break
            if not match_found:
                if pkg_release:
                    print('E: Release "%s" is unavailable for "%s"' % (pkg_release, pkg.name),
                          file=sys.stderr)
                else:
                    print('E: Version "%s" is unavailable for "%s"' % (pkg_version, pkg.name),
                          file=sys.stderr)
                self.close(14)
        else:
            _candidate = pkg.candidate
        candidate = _Package(
            version = _candidate.version,
            name = _candidate.package.name,
            fullname = None,
            architecture = pkg.architecture,
            source_name = _candidate.source_name,
            source_version = _candidate.source_version,
            uri = _candidate.uri,
            filename = os.path.basename(_candidate.filename),
            origin = _candidate.origins[0].origin,
            component = _candidate.origins[0].component,
            downloadable = _candidate.downloadable,
            is_installed = _candidate.is_installed,
            dependencies = _candidate.dependencies
        )
        if candidate.is_installed:
            candidate.installed_files = pkg.installed_files
        candidate.source_version_raw = candidate.source_version
        if ":" in candidate.source_version:
            candidate.source_version = candidate.source_version.split(":", 1)[1]
        return candidate

    def check_url(self, url:str, check_size:bool=True, stream:bool=True,
        msg:str=None):
        """ True if url can be downloaded and fits size requirements """
        if _DEBUG: print("Checking:", url)
        try:
            _r = requests.get(url, stream=stream, timeout=5)
        except Exception as e:
            _generic_exception_handler(e)
        else:
            if _r:
                if not _r.encoding:
                    _r.encoding = "utf-8"
                length = _r.headers.get("Content-Length")
                if length:
                    _r.length = int(length)
                else:
                    _r.length = 0
                if (not check_size or not
                    (check_size and _r.length > self.max_download_size and not
                    self.user_confirm(
                        (self.max_download_size_msg_lc if not msg else msg) %
                        (_r.length / self.MB))
                    )):
                    return _r
        if '_r' in vars():
            _r.close()
        return False

    @staticmethod
    def close(err:int=0):
        """ Exit """
        sys.exit(err)

    def exit_on_fail(self, err:int=404):
        """ Prints error message and calls self.close() """
        try:
            details = "Changelog unavailable for %s=%s" % (self.candidate.source_name, self.candidate.source_version_raw)
        except AttributeError:
            details = ""
        print("E: Failed to fetch changelog. %s" % details, file=sys.stderr)
        self.close(err)

    @staticmethod
    def strtobool (val):
        val = val.lower()
        if val in ('y', 'yes'):
            return True
        elif val in ('n', 'no'):
            return False
        else:
            raise ValueError("Invalid response value %s" % val)

    def user_confirm(self, q:str):
        """ returns bool (always False in non-interactive mode) """
        if not self.interactive:
            if _DEBUG: print("Maximum size exceeded, skipping in non-interactive mode")
            return False
        print("%s [y/n] " % q, end="")
        while True:
            try:
                response = self.strtobool(input())
                print("")
                return response
            except ValueError:
                print("Invalid response. Try again [y/n]: ", end="")
            except KeyboardInterrupt:
                pass

    def get_deb_or_tar(self, uri_tar:str=None):
        """ Returns request and URI of the preferred source

        The choice is made based on availability and size. If .deb is smaller
        than comparison_trigger_size, or if check_tar is False, then .deb is
        always selected.
        """
        comparison_trigger_size = 50000
        r_deb = self.check_url(self.candidate.uri, False)
        if r_deb:
            if uri_tar and r_deb.length > comparison_trigger_size:
                # try for .tar.xz
                r_tar = self.check_url(uri_tar, False)
                # validate and compare sizes
                if r_tar and r_tar.length < r_deb.length:
                    _r = r_tar
                    r_deb.close()
                else:
                    _r = r_deb
                    if r_tar:
                        r_tar.close()
            else:
                _r = r_deb
            if (not _r.length > self.max_download_size or
                self.user_confirm(self.max_download_size_msg_lc %
                (_r.length / self.MB))
                ):
                return (_r, _r.url)
        return (False, "")

    def get_changelog_from_filelist(self, filelist:list, local:bool=False):
        """ Returns hopefully the correct "changelog" or an empty string.

        We should not need to be searching because the debian policy says it
        must be at debian/changelog for source packages but not all seem to
        adhere to the policy:
        https://www.debian.org/doc/debian-policy/ch-source.html#debian-changelog-debian-changelog

        """
        files = [s for s in filelist if "changelog" in s.lower()]
        if local:
            testpath = "/usr/share/doc/%s/changelog" % self.candidate.name
            for item in files:
                if item.lower().startswith(testpath):
                    return item
        else:
            testpath = "debian/changelog"
            if testpath in files:
                return testpath
            testpath = "recipe/debian/changelog"
            if testpath in files:
                return testpath
            testpath = "usr/share/doc/%s/changelog" % self.candidate.name
            for item in files:
                if item.lower().startswith(testpath):
                    return item
            # no hits in the standard locations, let's try our luck in
            # random locations at the risk of getting the wrong file
            for item in files:
                if os.path.basename(item).lower().startswith("changelog"):
                    return item
        return None

    def get_apt_changelog_uri(self, uri_template:str):
        """ Returns URI based on provided apt changelog URI template.

        Emulates apt's std::string pkgAcqChangelog::URI
        The template must contain the @CHANGEPATH@ variable, which will
        be expanded to
            COMPONENT/SRC/SRCNAME/SRCNAME_SRCVER
        Component is omitted for releases without one (= flat-style
        repositories).
        """
        source_version = self.candidate.source_version

        def get_kernel_version_from_meta_package(pkg):
            for dependency in pkg.dependencies:
                if not dependency.target_versions or not dependency.rawtype == "Depends":
                    if _DEBUG: print("W: Kernel dependency not found:", dependency)
                    return None
                deppkg = dependency.target_versions[0]
                if deppkg.source_name in ("linux", "linux-signed"):
                    return deppkg.source_version
                if deppkg.source_name.startswith("linux-meta"):
                    _pkg = self.parse_package_metadata(str(deppkg))
                    return get_kernel_version_from_meta_package(_pkg)
            return None

        # Ubuntu kernel meta package workaround
        if self.candidate.origin == "Ubuntu" and \
           self.candidate.source_name.startswith("linux-meta"):
            _source_version = get_kernel_version_from_meta_package(self.candidate)
            if _source_version:
                source_version = _source_version
                self.candidate.source_name = "linux"

        # Ubuntu signed kernel workaround
        if self.candidate.origin == "Ubuntu" and \
           self.candidate.source_name == "linux-signed":
            self.candidate.source_name = "linux"

        # XXX:  Debian does not seem to reliably keep changelogs for previous
        #       (kernel) versions, so should we always look for the latest
        #       version instead on Debian? apt does not do this but the
        #       packages.debian.org website shows the latest version in the
        #       selected archive

        # strip epoch
        if ":" in source_version:
            source_version = source_version.split(":", 1)[1]

        # the path is: COMPONENT/SRC/SRCNAME/SRCNAME_SRCVER, e.g.
        #   main/a/apt/apt_1.1 or contrib/liba/libapt/libapt_2.0
        return uri_template.replace('@CHANGEPATH@',
            "%(component)s%(source_prefix)s/%(source_name)s/%(source_name)s_%(source_version)s" %
            {
                "component": self.candidate.component + "/" if \
                    self.candidate.component and \
                    self.candidate.component != "" else "",
                "source_prefix": self.source_prefix(),
                "source_name": self.candidate.source_name,
                "source_version": source_version
            })

    def source_prefix(self, source_name:str=None):
        """ Return prefix used for build repository URL """
        if not source_name:
            source_name = self.candidate.source_name
        return source_name[0] if not source_name.startswith("lib") else \
            source_name[:4]

    def parse_dsc(self, url:str):
        """ Returns filename or None """
        _r = self.check_url(url, False, False)
        if _r:
            target = ""
            lines = _r.text.split("Files:", 1)[1].split(":", 1)[0].split("-----BEGIN", 1)[0].split("\n")
            target = [s.strip() for s in lines if s.strip().lower().endswith('.debian.tar.xz')]
            if not target:
                target = [s.strip() for s in lines if s.strip().lower().endswith('.diff.gz')]
            if not target:
                target = [s.strip() for s in lines if s.strip().lower().endswith('.tar.xz')]
            # don't even test for .tar.gz, it will be too big compared to the .deb
            # if not target:
            #     target = [s.strip() for s in lines if s.strip().lower().endswith('.tar.gz')]
            if target:
                return target[0].split()[-1]
            elif _DEBUG: print(".dsc parse error for", url)
        return None

    def get_changelog_uri(self, base_uri:str):
        """ Tries to find a changelog in files listed in .dsc, locally cached
        packages as well as the remote .deb file

        Returns r and uri
        """
        uri = None
        # XXX:  For APT sources we could just read the apt_pkg.SourceRecords()
        #       directly, if available, which it is not for most users, so
        #       probably not worth it
        target_filename = self.parse_dsc("%s/%s_%s.dsc" % (base_uri, self.candidate.source_name, self.candidate.source_version))
        # get .debian.tar.xz or .diff.gz as a priority as the smallest options
        if (base_uri and target_filename and (
                target_filename.lower().endswith('.debian.tar.xz') or
                target_filename.lower().endswith('.diff.gz')
            )):
            uri = "%s/%s" % (base_uri, target_filename)
            target_filename = None
            r = self.check_url(uri, msg = self.max_download_size_msg)
        else:
            r = None
        if not r:
            # fall back to cached local package
            uri = self.apt_cache_path + self.candidate.filename
            if not os.path.isfile(uri):
                # cache miss, download the full source package or the .deb,
                # depending on size and availability
                if target_filename:
                    uri_tar = "%s/%s" % (base_uri, target_filename)
                else:
                    uri_tar = None
                r, uri = self.get_deb_or_tar(uri_tar)
        return (r, uri)

def _generic_exception_handler(e):
    if _DEBUG:
        import traceback
        print("%s: %s\n" % (e.__class__.__name__, traceback.format_exc()), file=sys.stderr)

def drop_cache():
    """ Drop the apt cache to free up memory. """
    # For some reason this does not free about 11M the first time it is run, and
    # additional 21M (32M total) the second time it runs (if the cache had been
    # opened again in the meantime). All consecutive runs keep those 32M total
    # without additional loss.
    # This only affects the freeing of memory, the maximum memory usage is stable.
    # It's unclear whether this is a python-apt or a python issue.

    if apt_changelog:
        apt_changelog.drop_cache()

def get_changelog(pkg_name:str, interactive:bool=False, output:bool=False,
    paged_output:bool=False, no_local:bool=False, max_download_size:int=0):
    """ Returns changelog for given package name, if any, and if within
        size-restrictions
    """
    changelog = None
    try:
        if not apt_changelog:
            __init__(interactive)
        if int(max_download_size) > 0:
            apt_changelog.max_download_size = int(max_download_size)
        else:
            apt_changelog.max_download_size = apt_changelog.max_download_size_default
        changelog = apt_changelog.get_changelog(pkg_name, no_local)
    except SystemExit:
        if interactive:
            raise
    except KeyboardInterrupt:
        sys.exit(130)
    else:
        if output:
            if not changelog:
                # empty changelog
                apt_changelog.exit_on_fail(7)
            if paged_output:
                try:
                    from pydoc import pager
                    pager(changelog)
                except Exception as e:
                    _generic_exception_handler(e)
                    paged_output = False
            else:
                print(changelog)
    return changelog

def print(*args, **kwargs):
    try:
        return __builtins__.print(*args, **kwargs)
    except:
        pass

def set_debug(value:bool):
    global _DEBUG
    _DEBUG = bool(value)

def __init__(interactive:bool=False):
    """ Instantiate _AptChangelog to global apt_changelog """
    global apt_changelog
    apt_changelog = _AptChangelog(interactive)

_DEBUG = False
apt_changelog = None

if __name__ == "__main__":
    if "--debug" in sys.argv:
        set_debug(True)
        sys.argv.remove("--debug")
    if "--no-local" in sys.argv:
        sys.argv.remove("--no-local")
        _no_local=True
    else:
        _no_local=False
    if len(sys.argv) != 2:
        print("""\
Usage:  apt changelog [options] <package>

Tries to retrieve the changelog of a package and display it through a pager.
By default it displays the changelog for the version that is installed.
However, you can specify the same options as for the install command.

Options:
        --no-local
            Always retrieve changelogs remotely, where possible. This can be
            useful when the locally installed changelog has been truncated.

Changelog lookup may fail for some packages if source repositories
are not enabled""")
        sys.exit(1)
    else:
        isatty = sys.stdin.isatty() and sys.stdout.isatty()
        get_changelog(sys.argv[1], interactive=isatty, output=True,
            paged_output=isatty, no_local=_no_local)



import time
import threading
import math
from pathlib import Path
import subprocess
import requests
import tempfile
import os

import gi
gi.require_version('AppStreamGlib', '1.0')
gi.require_version('Gtk', '3.0')
from gi.repository import AppStreamGlib, GLib, GObject, Gtk, Gio

try:
    gi.require_version('Flatpak', '1.0')
    from gi.repository import Flatpak
except:
    pass

from .pkgInfo import FlatpakPkgInfo
from . import dialogs
from .dialogs import ChangesConfirmDialog, FlatpakProgressWindow
from .misc import debug

class FlatpakRemoteInfo():
    def __init__(self, remote=None):
        if remote:
            self.name = remote.get_name()
            self.title = remote.get_title()
            self.url = remote.get_url()
            self.disabled = remote.get_disabled()
            self.noenumerate = remote.get_noenumerate()

            if not self.title or self.title == "":
                self.title = self.name.capitalize()
        else:
            self.name = None
            self.title = None
            self.url = None
            self.disabled = False
            self.noenumerate = False

    @classmethod
    def from_json(cls, json_data:dict):
        inst = cls()

        inst.name = json_data["name"]
        inst.title = json_data["title"]
        inst.url = json_data["url"]
        inst.disabled = json_data["disabled"]
        inst.noenumerate = json_data["noenumerate"]

        return inst

    def to_json(self):
        return self.__dict__

_fp_sys = None

_as_pool_lock = threading.Lock()
_as_pools = {} # keyed to remote name

def get_fp_sys():
    global _fp_sys

    if _fp_sys == None:
        _fp_sys = Flatpak.Installation.new_system(None)

    return _fp_sys

ALIASES = {
    "org.gnome.Weather" : "org.gnome.Weather.Application"
}

def make_pkg_hash(ref):
    if not isinstance(ref, Flatpak.Ref):
        raise TypeError("flatpak.make_pkg_hash() must receive FlatpakRef, not %s" % type(ref))

    try:
        return "fp:%s:%s" % (ref.get_origin(), ref.format_ref())
    except Exception:
        return "fp:%s:%s" % (ref.get_remote_name(), ref.format_ref())

def _get_remote_name_by_url(fp_sys, url):
    name = None

    try:
        remotes = fp_sys.list_remotes()
    except GLib.Error:
        remotes = []

    for remote in remotes:
        remote_url = remote.get_url()
        if remote_url.endswith('/'): #flatpakrefs are often missing the trailing forward slash in the url
            remote_url = remote_url[:-1]

        if remote_url == url:
            name = remote.get_name()

    return name

def _get_file_timestamp(gfile):
    try:
        info = gfile.query_info("time::modified", Gio.FileQueryInfoFlags.NONE, None)

        return info.get_attribute_uint64("time::modified")
    except GLib.Error as e:
        if e.code != Gio.IOErrorEnum.NOT_FOUND:
            print("Installer: flatpak - could not get time::modified from file %s" % gfile.get_path())
        return 0

def _should_update_appstream_data(fp_sys, remote, arch):
    ret = False

    gz_dir = remote.get_appstream_dir(arch)
    current_timestamp = _get_file_timestamp(gz_dir.get_child("appstream.xml.gz"))

    try:
        if fp_sys.update_remote_sync(remote.get_name(), None):
            print("Installer: flatpak - metadata for remote '%s' has been updated. Comparing appstream timestamps..." % remote.get_name())

            new_timestamp = _get_file_timestamp(remote.get_appstream_timestamp(arch))

            if (new_timestamp > current_timestamp) or (current_timestamp == 0):
                ret = True
    except GLib.Error as e:
        print("Installer: flatpak - could not update metadata for remote '%s': %s" % (remote.get_name(), e.message))

    return ret

def _process_remote(cache, fp_sys, remote, arch):
    remote_name = remote.get_name()

    if remote.get_disabled():
        print("Installer: flatpak - remote '%s' is disabled, skipping" % remote_name)
        return

    remote_url = remote.get_url()

    if _should_update_appstream_data(fp_sys, remote, arch):
        print("Installer: flatpak - new appstream data available for remote '%s', fetching..." % remote_name)

        try:
            fp_sys.update_appstream_sync(remote_name, arch, None)
        except GLib.Error:
            # Not fatal..
            pass
    else:
        print("Installer: flatpak - no new appstream data for remote '%s', skipping download" % remote_name)

    # get_noenumerate indicates whether a remote should be used to list applications.
    # Instead, they're intended for single downloads (via .flatpakref files)
    if remote.get_noenumerate():
        print("Installer: flatpak - remote '%s' is marked as no-enumerate skipping package listing" % remote_name)
        return

    try:
        for ref in fp_sys.list_remote_refs_sync(remote_name, None):
            if ref.get_kind() != Flatpak.RefKind.APP:
                continue

            if ref.get_arch() != arch:
                continue

            if ref.get_name().endswith("BaseApp"):
                continue

            _add_package_to_cache(cache, ref, remote_url, False)
    except GLib.Error as e:
        print(e.message)

def _add_package_to_cache(cache, ref, remote_url, installed):
    pkg_hash = make_pkg_hash(ref)

    try:
        remote_name = ref.get_remote_name()
    except Exception:
        remote_name = ref.get_origin()

    try:
        pkginfo = cache[pkg_hash]

        if installed:
            pkginfo.installed = installed
    except KeyError:
        pkginfo = FlatpakPkgInfo(pkg_hash, remote_name, ref, remote_url, installed)
        cache[pkg_hash] = pkginfo

    return pkginfo

def process_full_flatpak_installation(cache):
    fp_time = time.time()

    arch = Flatpak.get_default_arch()
    fp_sys = get_fp_sys()

    flatpak_remote_infos = {}

    try:
        for remote in fp_sys.list_remotes():
            _process_remote(cache, fp_sys, remote, arch)

            remote_name = remote.get_name()

            try:
                for ref in fp_sys.list_installed_refs_by_kind(Flatpak.RefKind.APP, None):
                    # All remotes will see installed refs, but the installed refs will always
                    # report their correct origin, so only add installed refs when they match the remote.
                    if ref.get_origin() == remote_name:
                        _add_package_to_cache(cache, ref, remote.get_url(), True)
            except GLib.Error as e:
                print(e.message)

            flatpak_remote_infos[remote_name] = FlatpakRemoteInfo(remote)

    except GLib.Error as e:
        print("Installer: flatpak - could not get remote list", e.message)
        cache = {}

    print('Installer: Processing Flatpaks for cache took %0.3f ms' % ((time.time() - fp_time) * 1000.0))

    return cache, flatpak_remote_infos

def _load_appstream_pool(pools, remote):
    pool = AppStreamGlib.Store()
    path = remote.get_appstream_dir().get_path()

    with open(os.path.join(path, "appstream.xml")) as f:
        pool.from_xml(f.read(), path)

    pools[remote.get_name()] = pool

def initialize_appstream():
    thread = threading.Thread(target=_initialize_appstream_thread)
    thread.start()

def _initialize_appstream_thread():
    fp_sys = get_fp_sys()

    global _as_pools
    global _as_pool_lock

    with _as_pool_lock:
        _as_pools = {}

        try:
            for remote in fp_sys.list_remotes():
                _load_appstream_pool(_as_pools, remote)
        except GLib.Error as e:
            print("Installer: Could not initialize appstream components for flatpaks: %s" % e.message)

def search_for_pkginfo_as_component(pkginfo):
    name = pkginfo.name

    comps = []

    global _as_pools
    global _as_pool_lock

    with _as_pool_lock:
        try:
            pool = _as_pools[pkginfo.remote]
        except Exception:
            return None

        comps = pool.get_apps_by_id(name + ".desktop")

        if comps == []:
            if name in ALIASES.keys():
                comps = pool.get_apps_by_id(ALIASES[name] + ".desktop")
            else:
                comps = pool.get_apps_by_id(name)
    if len(comps) > 0:
        return comps[0]
    else:
        return None

def _is_ref_installed(fp_sys, ref):
    try:
        iref = fp_sys.get_installed_ref(ref.get_kind(),
                                        ref.get_name(),
                                        ref.get_arch(),
                                        ref.get_branch(),
                                        None)

        if iref:
            return True
    except GLib.Error:
        pass
    except AttributeError: # bad/null ref
        pass

    return False

def _get_remote_sizes(fp_sys, remote, ref):
    # Since 0.11.4, this info is part of FlatpakRemoteRef
    # dl_s = ref.props.download_size
    # inst_s = ref.props.installed_size
    # But it appears to always contain identical dl and installed sizes,
    # so avoid it for now.

    try:
        success, dl_s, inst_s = fp_sys.fetch_remote_size_sync(remote,
                                                              ref,
                                                              None)
    except GLib.Error as e:
        print(e.message)
        # Not fatal?
        dl_s = 0
        inst_s = 0

    return dl_s, inst_s

def _get_installed_size(fp_sys, ref):
    if isinstance(ref, Flatpak.InstalledRef):
        return ref.get_installed_size()
    else:
        try:
            iref = fp_sys.get_installed_ref(ref.get_kind(),
                                            ref.get_name(),
                                            ref.get_arch(),
                                            ref.get_branch(),
                                            None)
            return iref.get_installed_size()
        except GLib.Error as e:
            print(e.message)
            # This isn't fatal I guess?
            return 0

def _add_to_list(ref_list, ref):
    ref_str = ref.format_ref()

    for existing_ref in ref_list:
        if ref_str == existing_ref.format_ref():
            debug("Skipping %s, already added to task" % ref_str)
            return

    ref_list.append(ref)

def _add_ref_to_task(fp_sys, task, ref, needs_update=False):
    if task.type == "install":
        if needs_update:
            _add_to_list(task.to_update, ref)
        else:
            _add_to_list(task.to_install, ref)

        dl_s, inst_s = _get_remote_sizes(fp_sys, ref.get_remote_name(), ref)

        task.download_size += dl_s
        task.install_size += inst_s
    elif task.type == "remove":
        _add_to_list(task.to_remove, ref)
        task.freed_size += _get_installed_size(fp_sys, ref)
    else:
        _add_to_list(task.to_update, ref)

        current_inst_s = _get_installed_size(fp_sys, ref)
        remote_dl_s, remote_inst_s = _get_remote_sizes(fp_sys, ref.get_remote_name(), ref)

        task.download_size += remote_dl_s

        if current_inst_s < remote_inst_s:
            task.install_size += remote_inst_s - current_inst_s
        else:
            task.freed_size += current_inst_s - remote_inst_s

def _find_remote_ref_from_list(fp_sys, remote_name, basic_ref, nofail=False):
    remote_ref = None

    all_refs = fp_sys.list_remote_refs_sync(remote_name, None)

    ref_str = basic_ref.format_ref()

    for ref in all_refs:
        if ref_str == ref.format_ref():
            remote_ref = ref
            break

    if remote_ref == None:
        try:
            remote_ref = fp_sys.fetch_remote_ref_sync(remote_name,
                                                      basic_ref.get_kind(),
                                                      basic_ref.get_name(),
                                                      basic_ref.get_arch(),
                                                      basic_ref.get_branch(),
                                                      None)
        except GLib.Error as e:
            if nofail:
                remote_ref = Flatpak.RemoteRef(remote_name=remote_name,
                                               kind=basic_ref.get_kind(),
                                               arch=basic_ref.get_arch(),
                                               branch=basic_ref.get_branch(),
                                               name=basic_ref.get_name(),
                                               commit=basic_ref.get_commit(),
                                               collection_id=basic_ref.get_collection_id())
    return remote_ref

def _get_runtime_ref(fp_sys, remote_name, ref):
    runtime_ref = None
    ref_string = None

    try:
        # Since 0.11.4 a FlatpakRemoteRef contains its metadata, so this
        # query is unnecessary.
        try:
            meta = ref.props.metadata

            if meta == None:
                raise AttributeError
        except AttributeError:
            meta = fp_sys.fetch_remote_metadata_sync(remote_name, ref, None)

        keyfile = GLib.KeyFile.new()

        data = meta.get_data().decode()

        keyfile.load_from_data(data, len(data), GLib.KeyFileFlags.NONE)

        runtime = keyfile.get_string("Application", "runtime")
        basic_ref = Flatpak.Ref.parse("runtime/%s" % runtime)

        ref_string = basic_ref.format_ref() # for error reporting only

        print("Looking for %s in %s" % (ref_string, remote_name))

        # prefer the same-remote's runtimes
        try:
            runtime_ref = _find_remote_ref_from_list(fp_sys, remote_name, basic_ref)

            if runtime_ref:
                print("Found runtime ref '%s' in remote %s" % (runtime_ref.format_ref(), remote_name))
        except GLib.Error:
            pass
        # if nothing is found, check other remotes
        if runtime_ref == None:
            for other_remote in fp_sys.list_remotes():
                other_remote_name = other_remote.get_name()

                if other_remote_name == remote_name:
                    continue

                print("Looking for %s in alternate remote %s" % (ref_string, other_remote_name))

                try:
                    runtime_ref = _find_remote_ref_from_list(fp_sys, other_remote_name, basic_ref)
                except GLib.Error:
                    continue

                if runtime_ref:
                    print("Found runtime ref '%s' in remote %s" % (runtime_ref.format_ref(), other_remote_name))
                    break
            if runtime_ref == None:
                raise Exception("Could not locate runtime '%s' in any registered remotes" % ref_string)
    except GLib.Error as e:
        runtime_ref = None
        raise Exception("Error finding runtimes for flatpak: %s" % e.message)

    return runtime_ref

def _get_remote_related_refs(fp_sys, remote_name, ref):
    return_refs = []

    try:
        related_refs = fp_sys.list_remote_related_refs_sync(remote_name,
                                                            ref.format_ref(),
                                                            None)

        for related_ref in related_refs:
            if not related_ref.should_download():
                continue

            print("Found related ref '%s' in remote %s" % (related_ref.format_ref(), remote_name))

            # Convert to a RemoteRef so that later functions know what remote
            # this is from.  Related refs should never fail from the given remote,
            # but if they're no-enumerate (as .flatpakref installs can be) then a listing
            # will be empty, so we have ..._from_list construct a synthetic one (nofail).
            remote_ref = _find_remote_ref_from_list(fp_sys, remote_name, related_ref, nofail=True)

            return_refs.append(remote_ref)
    except GLib.Error as e:
        raise Exception("Could not determine remote related refs for app: %s" % e.message)

    return return_refs

def _get_theme_refs(fp_sys, remote_name, ref=None):
    if ref:
        arch = ref.get_arch()
    else:
        arch = Flatpak.get_default_arch()

    theme_refs = []

    gtksettings = Gtk.Settings.get_default()

    icon_theme = "org.freedesktop.Platform.Icontheme.%s" % gtksettings.props.gtk_icon_theme_name
    gtk_theme = "org.gtk.Gtk3theme.%s" % gtksettings.props.gtk_theme_name

    def sortref(ref):
        try:
            val = float(ref.get_branch())
        except ValueError:
            val = 9.9

        return val

    for name in (icon_theme, gtk_theme):
        theme_ref = None

        try:
            print("Looking for theme %s in %s" % (name, remote_name))

            all_refs = fp_sys.list_remote_refs_sync(remote_name, None)

            matching_refs = []

            for listed_ref in all_refs:
                if listed_ref.get_name() == name:
                    matching_refs.append(listed_ref)

            if not matching_refs:
                continue

            # Sort highest version first.
            matching_refs = sorted(matching_refs, key=sortref, reverse=True)

            for matching_ref in matching_refs:
                if matching_ref.get_arch() != arch:
                    continue

                theme_ref = matching_ref
                print("Found theme ref '%s' in remote %s" % (theme_ref.format_ref(), remote_name))
                break

            # if nothing is found, check other remotes
            if theme_ref == None:
                for other_remote in fp_sys.list_remotes():
                    other_remote_name = other_remote.get_name()

                    if other_remote_name == remote_name:
                        continue

                    print("Looking for theme %s in alternate remote %s" % (name, other_remote_name))

                    all_refs = fp_sys.list_remote_refs_sync(other_remote_name, None)

                    matching_refs = []

                    for listed_ref in all_refs:
                        if listed_ref.get_name() == name:
                            matching_refs.append(listed_ref)

                    if not matching_refs:
                        continue

                    # Sort highest version first.
                    matching_refs = sorted(matching_refs, key=sortref, reverse=True)

                    for matching_ref in matching_refs:
                        if matching_ref.get_arch() != arch:
                            continue

                        theme_ref = matching_ref
                        print("Found theme ref '%s' in alternate remote %s" % (theme_ref.format_ref(), other_remote_name))
                        break

                    if theme_ref:
                        break
                if theme_ref == None:
                    debug("Could not locate theme '%s' in any registered remotes" % name)
        except GLib.Error as e:
            theme_ref = None
            debug("Error finding themes for flatpak: %s" % e.message)

        if theme_ref:
            theme_refs.append(theme_ref)

    return theme_refs

def get_updated_theme_refs():
    fp_sys = get_fp_sys()

    if not fp_sys.list_installed_refs_by_kind(Flatpak.RefKind.APP, None):
        return []

    theme_refs = []

    for remote in fp_sys.list_remotes():
        theme_refs += _get_theme_refs (fp_sys, remote.get_name())

    return theme_refs


def _get_installed_related_refs(fp_sys, remote, ref):
    return_refs = []

    try:
        related_refs = fp_sys.list_installed_related_refs_sync(remote,
                                                               ref.format_ref(),
                                                               None)
    except GLib.Error as e:
        raise Exception("Could not determine installed refs for app: %s" % e.message)

    return related_refs

def select_updates(task):
    thread = threading.Thread(target=_select_updates_thread, args=(task,))
    thread.start()

def _select_updates_thread(task):
    fp_sys = get_fp_sys()

    try:
        updates = fp_sys.list_installed_refs_for_update(None)
    except GLib.Error as e:
        task.info_ready_status = task.STATUS_BROKEN
        task.error_message = str(e)
        dialogs.show_flatpak_error(task.error_message)
        if task.info_ready_callback:
            GObject.idle_add(task.info_ready_callback, task)
        return

    for ref in updates:
        _add_ref_to_task(fp_sys, task, ref, needs_update=True)

    if len(task.to_update) > 0:
        print("flatpaks that can be updated:")
        for ref in task.to_update:
            print(ref.format_ref())

        task.info_ready_status = task.STATUS_OK
        task.execute = execute_transaction
    else:
        print("no updated flatpaks")

    if task.info_ready_callback:
        GObject.idle_add(task.info_ready_callback, task)

def select_packages(task):
    method = None

    if task.type == "install":
        method = _pick_refs_for_installation
    else:
        method = _pick_refs_for_removal

    print("Installer: Calculating changes required for Flatpak package: %s" % task.pkginfo.name)

    thread = threading.Thread(target=method, args=(task,))
    thread.start()

def _pick_refs_for_installation(task):
    fp_sys = get_fp_sys()

    pkginfo = task.pkginfo
    refid = pkginfo.refid

    ref = Flatpak.Ref.parse(refid)
    remote_name = pkginfo.remote

    try:
        remote_ref = _find_remote_ref_from_list(fp_sys, remote_name, ref, nofail=True)
        print("Selected ref to install: '%s' in remote %s" % (remote_ref.format_ref(), remote_name))

        _add_ref_to_task(fp_sys, task, remote_ref)
        update_list = fp_sys.list_installed_refs_for_update(None)

        runtime_ref = _get_runtime_ref(fp_sys, remote_name, remote_ref)

        if not _is_ref_installed(fp_sys, runtime_ref):
            _add_ref_to_task(fp_sys, task, runtime_ref)
        else:
            if runtime_ref in update_list:
                _add_ref_to_task(fp_sys, task, runtime_ref, needs_update=True)

        all_related_refs = _get_remote_related_refs(fp_sys, remote_ref.get_remote_name(), remote_ref)
        all_related_refs += _get_remote_related_refs(fp_sys, runtime_ref.get_remote_name(), runtime_ref)
        all_related_refs += _get_theme_refs(fp_sys, runtime_ref.get_remote_name(), runtime_ref)

        for related_ref in all_related_refs:
            if not _is_ref_installed(fp_sys, related_ref):
                _add_ref_to_task(fp_sys, task, related_ref)
            else:
                if related_ref in update_list:
                    _add_ref_to_task(fp_sys, task, related_ref, needs_update=True)

    except Exception as e:
        # Something went wrong, bail out
        task.info_ready_status = task.STATUS_BROKEN
        task.error_message = str(e)
        dialogs.show_flatpak_error(task.error_message)
        if task.info_ready_callback:
            GObject.idle_add(task.info_ready_callback, task)
        return

    print("For installation:")
    for ref in task.to_install:
        print(ref.format_ref())

    task.info_ready_status = task.STATUS_OK
    task.execute = execute_transaction

    if task.info_ready_callback:
        GObject.idle_add(task.info_ready_callback, task)

def _pick_refs_for_removal(task):
    fp_sys = get_fp_sys()

    pkginfo = task.pkginfo

    try:
        ref = fp_sys.get_installed_ref(pkginfo.kind,
                                       pkginfo.name,
                                       pkginfo.arch,
                                       pkginfo.branch,
                                       None)

        remote = ref.get_origin()

        _add_ref_to_task(fp_sys, task, ref)

        related_refs = _get_installed_related_refs(fp_sys, remote, ref)

        for related_ref in related_refs:
            if _is_ref_installed(fp_sys, related_ref) and related_ref.should_delete():
                _add_ref_to_task(fp_sys, task, related_ref)

    except Exception as e:
        task.info_ready_status = task.STATUS_BROKEN
        task.error_message = str(e)
        dialogs.show_flatpak_error(task.error_message)
        GObject.idle_add(task.info_ready_callback, task)
        return

    print("For removal:")
    for ref in task.to_remove:
        print(ref.format_ref())

    task.info_ready_status = task.STATUS_OK
    task.execute = execute_transaction

    if task.info_ready_callback:
        GObject.idle_add(task.info_ready_callback, task)

def list_updated_pkginfos(cache):
    fp_sys = get_fp_sys()

    updated = []

    try:
        updates = fp_sys.list_installed_refs_for_update(None)
    except GLib.Error as e:
        print("Installer: flatpak - could not get updated flatpak refs")
        return []

    for ref in updates:
        pkg_hash = make_pkg_hash(ref)

        try:
            updated.append(cache[pkg_hash])
        except KeyError:
            pass

    return updated

def find_pkginfo(cache, string):
    for key in cache.get_subset_of_type("f").keys():
        candidate = cache[key]

        if string == candidate.name:
            return candidate

    return None

def generate_uncached_pkginfos(cache):
    fp_sys = get_fp_sys()

    try:
        for remote in fp_sys.list_remotes():
            remote_name = remote.get_name()

            for ref in fp_sys.list_installed_refs_by_kind(Flatpak.RefKind.APP, None):
                # All remotes will see installed refs, but the installed refs will always
                # report their correct origin, so only add installed refs when they match the remote.
                if ref.get_origin() == remote_name:
                    _add_package_to_cache(cache, ref, remote.get_url(), True)

    except GLib.Error as e:
        print("Installer: flatpak - could not check for uncached pkginfos", e.message)

def pkginfo_is_installed(pkginfo):
    fp_sys = get_fp_sys()

    try:
        iref = fp_sys.get_installed_ref(pkginfo.kind,
                                        pkginfo.name,
                                        pkginfo.arch,
                                        pkginfo.branch,
                                        None)

        if iref:
            return True
    except GLib.Error:
        pass

    return False

def list_remotes():
    fp_sys = get_fp_sys()

    remotes = []

    try:
        for remote in fp_sys.list_remotes():
            remotes.append(FlatpakRemoteInfo(remote))

    except GLib.Error as e:
        print("Installer: flatpak - could not fetch remote list", e.message)
        remotes = []

    return remotes

def get_pkginfo_from_file(cache, file, callback):
    thread = threading.Thread(target=_pkginfo_from_file_thread, args=(cache, file, callback))
    thread.start()

def _pkginfo_from_file_thread(cache, file, callback):
    fp_sys = get_fp_sys()

    path = file.get_path()

    if path == None:
        print("Installer: flatpak - no valid .flatpakref path provided")
        return None

    ref = None
    pkginfo = None
    remote_name = None

    with open(path) as f:
        contents = f.read()

        b = contents.encode("utf-8")
        gb = GLib.Bytes(b)

        try:
            kf = GLib.KeyFile()
            if kf.load_from_file(path, GLib.KeyFileFlags.NONE):
                name = kf.get_string("Flatpak Ref", "Name")
                url = kf.get_string("Flatpak Ref", "Url")

                try:
                    branch = kf.get_string("Flatpak Ref", "Branch")
                except GLib.Error as e:
                    if e.code == GLib.KeyFileError.KEY_NOT_FOUND:
                        print("Installer: flatpak - flatpakref file doesn't have a Branch key, maybe nightly or testing.")
                        branch = None

                remote_name = _get_remote_name_by_url(fp_sys, url)

                if name and remote_name:
                    ref = Flatpak.RemoteRef(remote_name=remote_name,
                                            kind=Flatpak.RefKind.APP,
                                            arch=Flatpak.get_default_arch(),
                                            branch=branch,
                                            name=name)
                    print("Installer: flatpak - using existing remote '%s' for flatpakref file install" % remote_name)
                else: #If Flatpakref is not installed already
                    try:
                        print("Installer: flatpak - trying to install new remote for flatpakref file")
                        ref = fp_sys.install_ref_file(gb, None)
                        fp_sys.drop_caches(None)

                        remote_name = ref.get_remote_name()
                        print("Installer: flatpak - added remote '%s'" % remote_name)
                    except GLib.Error as e:
                        if e.code != Gio.DBusError.ACCESS_DENIED: # user cancelling auth prompt for adding a remote
                            print("Installer: could not add new remote to system: %s" % e.message)
                            dialogs.show_flatpak_error(e.message)
        except GLib.Error as e:
            print("Installer: flatpak - could not parse flatpakref file: %s" % e.message)
            dialogs.show_flatpak_error(e.message)

        if ref:
            try:
                remote = fp_sys.get_remote_by_name(remote_name, None)

                # Add the ref's remote if it doesn't already exist
                _process_remote(cache, fp_sys, remote, Flatpak.get_default_arch())

                # Add the ref to the cache, so we can work with it like any other in mintinstall
                pkginfo = _add_package_to_cache(cache, ref, remote.get_url(), False)

                # Fetch the appstream info for the ref
                global _as_pools

                with _as_pool_lock:
                    if remote_name not in _as_pools.keys():
                        _load_appstream_pool(_as_pools, remote)

                # Some flatpakref files will have a pointer to a runtime .flatpakrepo file
                # We need to process and possibly add that remote as well.

                kf = GLib.KeyFile()
                if kf.load_from_file(path, GLib.KeyFileFlags.NONE):
                    try:
                        url = kf.get_string("Flatpak Ref", "RuntimeRepo")
                    except GLib.Error:
                        url = None

                    if url:
                        # Fetch the .flatpakrepo file
                        r = requests.get(url, stream=True)

                        file = tempfile.NamedTemporaryFile(delete=False)

                        with file as fd:
                            for chunk in r.iter_content(chunk_size=128):
                                fd.write(chunk)

                        # Get the true runtime url from the repo file
                        runtime_repo_url = _get_repofile_repo_url(file.name)

                        if runtime_repo_url:
                            existing = False

                            path = Path(file.name)
                            runtime_remote_name = Path(url).stem

                            # Check if the remote is already installed
                            for remote in fp_sys.list_remotes(None):
                                # See comments below in _remote_from_repo_file_thread about get_noenumerate() use.
                                if remote.get_url() == runtime_repo_url and not remote.get_noenumerate():
                                    print("Installer: flatpak - runtime remote '%s' already in system, skipping" % runtime_remote_name)
                                    existing = True
                                    break

                            if not existing:
                                print("Installer: Adding additional runtime remote named '%s' at '%s'" % (runtime_remote_name, runtime_repo_url))

                                cmd_v = ['flatpak',
                                         'remote-add',
                                         '--from',
                                         runtime_remote_name,
                                         file.name]

                                add_repo_proc = subprocess.Popen(cmd_v)
                                retcode = add_repo_proc.wait()

                                fp_sys.drop_caches(None)
                        os.unlink(file.name)
            except GLib.Error as e:
                print("Installer: could not process .flatpakref file: %s" % e.message)
                dialogs.show_flatpak_error(e.message)

    GLib.idle_add(callback, pkginfo, priority=GLib.PRIORITY_DEFAULT)

def add_remote_from_repo_file(cache, file, callback):
    thread = threading.Thread(target=_remote_from_repo_file_thread, args=(cache, file, callback))
    thread.start()

def _remote_from_repo_file_thread(cache, file, callback):
    try:
        path = Path(file.get_path())
    except TypeError:
        print("Installer: flatpak - no valid .flatpakrepo path provided")
        return

    fp_sys = get_fp_sys()

    # Make sure the remote isn't already setup (even under a different name)
    # We need to exclude -origin repos - they're added for .flatpakref files
    # if the remote isn't installed already, and get auto-removed when the app
    # the ref file describes gets uninstalled.  -origin remotes are also marked
    # no-enumerate, so we can filter them here by that.

    existing = False

    url = _get_repofile_repo_url(path)

    if url:
        for remote in fp_sys.list_remotes(None):
            if remote.get_url() == url and not remote.get_noenumerate():
                existing = True
                break

    if existing:
        GObject.idle_add(callback, file, "exists")
        return

    cmd_v = ['flatpak',
             'remote-add',
             '--from',
             path.stem,
             path]

    add_repo_proc = subprocess.Popen(cmd_v, stderr=subprocess.PIPE)
    stdout, stderr = add_repo_proc.communicate()

    if "Error.AccessDenied" in stderr.decode():
        GObject.idle_add(callback, file, "cancel")
        return

    if add_repo_proc.returncode != 0 and "already exists" not in stderr.decode():
        GObject.idle_add(callback, file, "error")
        return

    # We'll do a full cache rebuild - otherwise, after this installer session, the
    # new apps from this remote won't show up until the next scheduled cache rebuild.
    try:
        fp_sys.drop_caches(None)
    except GLib.Error:
        pass

    cache.force_new_cache_async(callback)

def _get_repofile_repo_url(path):
    kf = GLib.KeyFile()

    try:
        if kf.load_from_file(str(path), GLib.KeyFileFlags.NONE):
            url = kf.get_string("Flatpak Repo", "Url")

            return url
    except GLib.Error as e:
        print(e.message)

    return None

def execute_transaction(task):
    if len(task.to_install + task.to_remove + task.to_update) > 1:
        dia = ChangesConfirmDialog(None, task, parent=task.parent_window)
        res = dia.run()
        dia.hide()
        if res != Gtk.ResponseType.OK:
            GObject.idle_add(task.finished_cleanup_cb, task)
            return

    if task.client_progress_cb != None:
        task.has_window = True
    else:
        progress_window = FlatpakProgressWindow(task)
        progress_window.present()

    thread = threading.Thread(target=_execute_transaction_thread, args=(task,))
    thread.start()

def _execute_transaction_thread(task):
    GLib.idle_add(task.client_progress_cb, task.pkginfo, 0, True, priority=GLib.PRIORITY_DEFAULT)

    fp_sys = get_fp_sys()

    task.transaction = MetaTransaction(fp_sys, task)

class MetaTransaction():
    def __init__(self, fp_sys, task):
        self.fp_sys = fp_sys
        self.task = task
        pkginfo = self.pkginfo = task.pkginfo

        task.to_install.reverse()

        self.item_count = len(task.to_install + task.to_remove + task.to_update)
        self.current_count = 0

        for ref in task.to_install:
            task.progress_state = task.PROGRESS_STATE_INSTALLING
            task.current_package_name = ref.get_name()

            print("installing: %s" % ref.format_ref())
            try:
                self.fp_sys.install(ref.get_remote_name(),
                                    ref.get_kind(),
                                    ref.get_name(),
                                    ref.get_arch(),
                                    ref.get_branch(),
                                    self.on_flatpak_progress,
                                    None,
                                    task.cancellable)
            except GLib.Error as e:
                if task.current_package_name == task.pkginfo.name:
                    if e.code != Gio.IOErrorEnum.CANCELLED:
                        task.progress_state = task.PROGRESS_STATE_FAILED
                        task.current_package_name = None
                        self.on_flatpak_error(e.message)
                        return
                else:
                    print("Installer: flatpak - problem installing %s: %s" % (task.current_package_name, e.message))

            self.current_count += 1

        for ref in task.to_remove:
            task.progress_state = task.PROGRESS_STATE_REMOVING
            task.current_package_name = ref.get_name()

            print("removing: %s" % ref.format_ref())
            try:
                self.fp_sys.uninstall(ref.get_kind(),
                                      ref.get_name(),
                                      ref.get_arch(),
                                      ref.get_branch(),
                                      self.on_flatpak_progress,
                                      None,
                                      task.cancellable)
            except GLib.Error as e:
                if task.current_package_name == task.pkginfo.name:
                    if e.code != Gio.IOErrorEnum.CANCELLED:
                        task.progress_state = task.PROGRESS_STATE_FAILED
                        task.current_package_name = None
                        self.on_flatpak_error(e.message)
                        return
                else:
                    print("Installer: flatpak - problem removing %s: %s" % (task.current_package_name, e.message))

            self.current_count += 1

        for ref in task.to_update:
            task.progress_state = task.PROGRESS_STATE_UPDATING
            task.current_package_name = ref.get_name()

            print("updating: %s" % ref.format_ref())
            try:
                self.fp_sys.update(Flatpak.UpdateFlags.NONE,
                                   ref.get_remote_name(),
                                   ref.get_kind(),
                                   ref.get_name(),
                                   ref.get_arch(),
                                   ref.get_branch(),
                                   self.on_flatpak_progress,
                                   None,
                                   task.cancellable)
            except GLib.Error as e:
                # update will None as a task's pkginfo, just warn.
                print("Installer: flatpak - problem updating %s: %s" % (task.current_package_name, e.message))

            self.current_count += 1

        task.progress_state = task.PROGRESS_STATE_FINISHED
        task.current_package_name = None
        self.on_flatpak_finished()

    def on_flatpak_progress(self, status, progress, estimating, data=None):
        # Simple for now, each package gets an equal slice, and package progress is a percentage of that slice

        package_chunk_size = 1.0 / self.item_count
        partial_chunk = (progress / 100.0) * package_chunk_size

        actual_progress = math.floor(((self.current_count * package_chunk_size) + partial_chunk) * 100.0)

        GLib.idle_add(self.task.client_progress_cb,
                      self.task.pkginfo,
                      actual_progress,
                      estimating,
                      priority=GLib.PRIORITY_DEFAULT)

    def on_flatpak_error(self, error_details):
        self.task.error_message = error_details

        # Show an error popup only if we're in mintinstall, otherwise a flatpak
        # progress window will show the error details
        if self.task.has_window:
            dialogs.show_flatpak_error(error_details)

        GLib.idle_add(self.task.error_cleanup_cb, self.task)

    def on_flatpak_finished(self):
        self.fp_sys.drop_caches(None)

        GLib.idle_add(self.task.finished_cleanup_cb, self.task)





import gi
gi.require_version('XApp', '1.0')
from gi.repository import GLib, Gtk, GObject, Gdk, XApp

import gettext
APP = 'mint-common'
LOCALE_DIR = "/usr/share/linuxmint/locale"
t = gettext.translation(APP, LOCALE_DIR, fallback=True)
_ = t.gettext

from aptdaemon.gtk3widgets import AptConfirmDialog

######################### Subclass Apt's dialog to keep consistency

class ChangesConfirmDialog(AptConfirmDialog):

    """Dialog to confirm the changes that would be required by a
    transaction.
    """

    def __init__(self, transaction, task=None, parent=None):
        super(ChangesConfirmDialog, self).__init__(transaction, cache=None, parent=parent)
        self.parent_window = parent
        self.set_size_request(500, 350)
        self.task = task

    def _show_changes(self):
        """Show a message and the dependencies in the dialog."""
        self.treestore.clear()

        if not self.parent_window:
            self.set_skip_taskbar_hint(True)
            self.set_keep_above(True)

        # Run parent method for apt
        if not self.task or self.task.pkginfo.pkg_hash.startswith("a"):
            super(ChangesConfirmDialog, self)._show_changes()
        else:
            # flatpak
            self.set_title(_("Flatpaks"))

            if len(self.task.to_install) > 0:
                piter = self.treestore.append(None, ["<b>%s</b>" % _("Install")])

                for ref in self.task.to_install:
                    if self.task.pkginfo.refid == ref.format_ref():
                        continue

                    self.treestore.append(piter, [ref.get_name()])

            if len(self.task.to_remove) > 0:
                piter = self.treestore.append(None, ["<b>%s</b>" % _("Remove")])

                for ref in self.task.to_remove:
                    if self.task.pkginfo.refid == ref.format_ref():
                        continue

                    self.treestore.append(piter, [ref.get_name()])

            if len(self.task.to_update) > 0:
                piter = self.treestore.append(None, ["<b>%s</b>" % _("Upgrade")])

                for ref in self.task.to_update:
                    if self.task.pkginfo.refid == ref.format_ref():
                        continue

                    self.treestore.append(piter, [ref.get_name()])

            msg = _("Please take a look at the list of changes below.")

            if len(self.treestore) == 1:
                filtered_store = self.treestore.filter_new(
                    Gtk.TreePath.new_first())
                self.treeview.expand_all()
                self.treeview.set_model(filtered_store)
                self.treeview.set_show_expanders(False)

                if len(self.task.to_install) > 0:
                    title = _("Additional software has to be installed")
                elif len(self.task.to_remove) > 0:
                    title = _("Additional software has to be removed")
                elif len(self.task.to_update) > 0:
                    title = _("Additional software has to be upgraded")

                if len(filtered_store) < 6:
                    self.set_resizable(False)
                    self.scrolled.set_policy(Gtk.PolicyType.AUTOMATIC,
                                             Gtk.PolicyType.NEVER)
                else:
                    self.treeview.set_size_request(350, 200)
            else:
                title = _("Additional changes are required")
                self.treeview.set_size_request(350, 200)
                self.treeview.collapse_all()

            if self.task.download_size > 0:
                msg += "\n"
                msg += (_("%s will be downloaded in total.") %
                        GLib.format_size(self.task.download_size))
            if self.task.freed_size > 0:
                msg += "\n"
                msg += (_("%s of disk space will be freed.") %
                        GLib.format_size(self.task.freed_size))
            elif self.task.install_size > 0:
                msg += "\n"
                msg += (_("%s more disk space will be used.") %
                        GLib.format_size(self.task.install_size))
            self.label.set_markup("<b><big>%s</big></b>\n\n%s" % (title, msg))

    def render_package_desc(self, column, cell, model, iter, data):
        value = model.get_value(iter, 0)

        cell.set_property("markup", value)


class FlatpakProgressWindow(Gtk.Dialog):
    """
    Progress dialog for standalone flatpak installs, removals, updates.
    Intended to be used when not working as part of a parent app (like mintinstall)
    """

    def __init__(self, task, parent=None):
        Gtk.Dialog.__init__(self, parent=parent)

        self.task = task
        self.finished = False

        # Progress goes directly to this window
        task.client_progress_cb = self.window_client_progress_cb

        # finished callbacks route thru the installer
        # but we want to see them in this window also.
        self.final_finished_cb = task.client_finished_cb
        task.client_finished_cb = self.window_client_finished_cb

        self.pulse_timer = 0
        self.active_task_state = task.progress_state

        self.real_progress_text = None
        self.num_dots = 0

        # Setup the dialog
        self.set_border_width(6)
        self.set_resizable(False)
        self.get_content_area().set_spacing(6)
        # Setup the cancel button
        self.button = Gtk.Button.new_from_stock(Gtk.STOCK_CANCEL)
        self.button.set_use_stock(True)
        self.get_action_area().pack_start(self.button, False, False, 0)
        self.button.connect("clicked", self.on_button_clicked)
        self.button.show()

        # labels and progressbar
        hbox = Gtk.HBox()
        hbox.set_spacing(12)
        hbox.set_border_width(6)
        vbox = Gtk.VBox()
        vbox.set_spacing(12)

        self.phase_label = Gtk.Label()
        vbox.pack_start(self.phase_label, False, False, 0)
        self.phase_label.set_halign(Gtk.Align.START)

        vbox_progress = Gtk.VBox()
        vbox_progress.set_spacing(6)
        self.progress = Gtk.ProgressBar()
        vbox_progress.pack_start(self.progress, False, True, 0)

        self.progress_label = Gtk.Label()
        vbox_progress.pack_start(self.progress_label, False, False, 0)
        self.progress_label.set_halign(Gtk.Align.START)
        self.progress_label.set_line_wrap(True)
        self.progress_label.set_max_width_chars(60)

        vbox.pack_start(vbox_progress, False, True, 0)
        hbox.pack_start(vbox, True, True, 0)

        self.get_content_area().pack_start(hbox, True, True, 0)

        self.set_title(_("Flatpak Progress"))
        XApp.set_window_icon_name(self, "system-software-installer")

        hbox.show_all()
        self.realize()

        self.progress.set_size_request(350, -1)
        functions = Gdk.WMFunction.MOVE | Gdk.WMFunction.RESIZE
        try:
            self.get_window().set_functions(functions)
        except TypeError:
            # workaround for older and broken GTK typelibs
            self.get_window().set_functions(Gdk.WMFunction(functions))

        self.update_labels()

        # catch ESC and behave as if cancel was clicked
        self.connect("delete-event", self._on_dialog_delete_event)

    def set_progress_text(self, text):
        if text != self.real_progress_text:
            self.real_progress_text = text
            self.progress_label.set_text(text)
        else:
            self.tick()

    def tick(self):
        if self.num_dots < 5:
            self.num_dots += 1
        else:
            self.num_dots = 0

        new_string = self.real_progress_text

        i = 0

        while i < self.num_dots:
            new_string += "."
            i += 1

        self.progress_label.set_text(new_string)

    def update_labels(self):
        phase_label = ""

        if self.task.cancellable.is_cancelled():
            phase_label = _("Cancelled")
            self.set_progress_text(_("The operation was cancelled before it could complete."))
        elif self.task.progress_state == self.task.PROGRESS_STATE_INIT:
            phase_label = _("Initializing")
            self.set_progress_text("")
        elif self.task.progress_state == self.task.PROGRESS_STATE_INSTALLING:
            phase_label = _("Installing flatpaks")
            self.set_progress_text(_("Installing: %s") % self.task.current_package_name)
        elif self.task.progress_state == self.task.PROGRESS_STATE_REMOVING:
            phase_label = _("Removing flatpaks")
            self.set_progress_text(_("Removing: %s") % self.task.current_package_name)
        elif self.task.progress_state == self.task.PROGRESS_STATE_UPDATING:
            phase_label = _("Updating flatpaks")
            self.set_progress_text(_("Updating: %s") % self.task.current_package_name)
        elif self.task.progress_state == self.task.PROGRESS_STATE_FINISHED:
            phase_label = _("Finished")
            self.set_progress_text(_("Operation completed successfully"))
        elif self.task.progress_state == self.task.PROGRESS_STATE_FAILED:
            phase_label = _("Failed")
            self.set_progress_text(_("An error occurred:\n\n%s") % self.task.error_message)

        self.phase_label.set_markup("<big><b>%s</b></big>" % phase_label)
        self.set_title(phase_label)

    def start_progress_pulse(self):
        if self.pulse_timer > 0:
            return

        self.progress.pulse()
        self.pulse_timer = GObject.timeout_add(1050, self.progress_pulse_tick)

    def progress_pulse_tick(self):
        self.progress.pulse()

        return GLib.SOURCE_CONTINUE

    def stop_progress_pulse(self):
        if self.pulse_timer > 0:
            GObject.source_remove(self.pulse_timer)
            self.pulse_timer = 0

    def _on_dialog_delete_event(self, dialog, event):
        self.button.clicked()
        return True

    def window_client_progress_cb(self, pkginfo, progress, estimating):
        if estimating:
            self.start_progress_pulse()
        else:
            self.stop_progress_pulse()

            self.progress.set_fraction(progress / 100.0)
            XApp.set_window_progress(self, progress)
            self.tick()

        self.update_labels()

    def window_client_finished_cb(self, pkginfo, error_message):
        self.finished = True

        XApp.set_window_progress(self, 0)
        self.stop_progress_pulse()

        self.progress.set_fraction(1.0)

        self.update_labels()

        if error_message:
            self.set_urgency_hint(True)
            self.button.set_label(Gtk.STOCK_CLOSE)
        else:
            self.destroy()
            self.final_finished_cb(self.task.pkginfo, error_message)

    def on_button_clicked(self, button):
        if not self.finished:
            self.task.cancellable.cancel()
        else:
            self.destroy()
            self.final_finished_cb(self.task.pkginfo, self.task.error_message)

def show_flatpak_error(message):
    GObject.idle_add(_show_flatpak_error_mainloop, message, priority=GLib.PRIORITY_DEFAULT)

def _show_flatpak_error_mainloop(message):

    dialog = Gtk.MessageDialog(None,
                               Gtk.DialogFlags.DESTROY_WITH_PARENT,
                               Gtk.MessageType.ERROR,
                               Gtk.ButtonsType.OK,
                               "")

    text = _("An error occurred")
    dialog.set_markup("<big><b>%s</b></big>" % text)
    message_label = Gtk.Label(message)
    dialog.get_message_area().pack_start(message_label, False, False, 6)
    message_label.set_line_wrap(True)
    message_label.set_max_width_chars(60)
    message_label.show()

    dialog.run()
    dialog.hide()

    return False



import time
import threading
import apt

import gi
gi.require_version('AppStream', '1.0')
gi.require_version('Gtk', '3.0')
gi.require_version('Gdk', '3.0')
from gi.repository import GObject, Gtk

import aptdaemon.client
from aptdaemon.gtk3widgets import AptErrorDialog, AptProgressDialog
import aptdaemon.errors

from .pkgInfo import AptPkgInfo
from .dialogs import ChangesConfirmDialog

# List of packages which are either broken or do not install properly in mintinstall
BROKEN_PACKAGES = ['pepperflashplugin-nonfree']

# List extra packages that aren't necessarily marked in their control files, but
# we know better...
CRITICAL_PACKAGES = ["mint-common", "mint-meta-core", "mintdesktop"]

def capitalize(string):
    if len(string) > 1:
        return (string[0].upper() + string[1:])

    return (string)

_apt_cache = None
_apt_cache_lock = threading.Lock()
_as_pool = None

def get_apt_cache(full=False):
    global _apt_cache

    if full or (not _apt_cache):
        with _apt_cache_lock:
            _apt_cache = apt.Cache()

    return _apt_cache

def add_prefix(name):
    return "apt:%s" % (name)

def make_pkg_hash(apt_pkg):
    if not isinstance(apt_pkg, apt.Package):
        raise TypeError("apt.make_pkg_hash_make must receive apt.Package, not %s" % type(apt_pkg))

    return add_prefix(apt_pkg.name)

def process_full_apt_cache(cache):
    apt_time = time.time()
    apt_cache = get_apt_cache()

    sections = {}

    keys = apt_cache.keys()

    for key in keys:
        name = apt_cache[key].name
        pkg = apt_cache[key]

        if name.startswith("lib") and not name.startswith(("libreoffice", "librecad", "libk3b7", "libimage-exiftool-perl")):
            continue
        if name.endswith(":i386") and name != "steam:i386":
            continue
        if name.endswith("-dev"):
            continue
        if name.endswith("-dbg"):
            continue
        if name.endswith("-doc"):
            continue
        if name.endswith("-common"):
            continue
        if name.endswith("-data"):
            continue
        if "-locale-" in name:
            continue
        if "-l10n-" in name:
            continue
        if name.endswith("-dbgsym"):
            continue
        if name.endswith("l10n"):
            continue
        if name.endswith("-perl"):
            continue
        if name == "snapd":
            continue
        if ":" in name and name.split(":")[0] in keys:
            continue
        try:
            if "transitional" in pkg.candidate.summary.lower():
                continue
        except Exception as e:
            print("Problem parsing package (maybe it's virtual): %s: %s" % (name, e))
            continue
            # pass

        pkg_hash = make_pkg_hash(pkg)

        if pkg.candidate:
            section_string = pkg.candidate.section

            if "/" in section_string:
                section = section_string.split("/")[1]
            else:
                section = section_string

        try:
            sections[section].append(pkg_hash)
        except Exception:
            sections[section] = []
            sections[section].append(pkg_hash)

        cache[pkg_hash] = AptPkgInfo(pkg_hash, pkg)

    print('Installer: Processing APT packages for cache took %0.3f ms' % ((time.time() - apt_time) * 1000.0))

    return cache, sections

# def initialize_appstream():
#     global _as_pool

#     if _as_pool == None:
#         pool = AppStream.Pool()
#         pool.set_cache_flags(AppStream.CacheFlags.NONE)
#         pool.load()
#         _as_pool = pool

def search_for_pkginfo_apt_pkg(pkginfo):
    name = pkginfo.name

    apt_cache = get_apt_cache()

    try:
        return apt_cache[name]
    except:
        return None

def find_pkginfo(cache, string):
    try:
        pkginfo = cache[add_prefix(string)]
    except:
        pkginfo = None

    return pkginfo

def pkginfo_is_installed(pkginfo):
    global _apt_cache_lock
    apt_cache = get_apt_cache()

    with _apt_cache_lock:
        try:
            return apt_cache[pkginfo.name].installed != None
        except:
            return False

def select_packages(task):
    thread = threading.Thread(target=_calculate_apt_changes, args=(task,))
    thread.start()

def _is_critical_package(pkg):
    try:
        if pkg.essential or pkg.versions[0].priority == "required" or pkg.name in CRITICAL_PACKAGES:
            return True

        return False
    except Exception:
        return False

def _calculate_apt_changes(task):
    global _apt_cache_lock
    apt_cache = get_apt_cache()

    with _apt_cache_lock:
        apt_cache.clear()

        print("Installer: Calculating changes required for APT package: %s" % task.pkginfo.name)

        pkginfo = task.pkginfo

        aptpkg = apt_cache[pkginfo.name]

        try:
            if aptpkg.is_installed:
                aptpkg.mark_delete(True, True)
            else:
                aptpkg.mark_install()
        except:
            if aptpkg.name not in BROKEN_PACKAGES:
                BROKEN_PACKAGES.append(aptpkg.name)

        changes = apt_cache.get_changes()

        for pkg in changes:
            if pkg.marked_install:
                task.to_install.append(pkg.name)
            elif pkg.marked_upgrade:
                task.to_update.append(pkg.name)
            elif pkg.marked_delete:
                task.to_remove.append(pkg.name)

        task.download_size = apt_cache.required_download

        space = apt_cache.required_space

        if space < 0:
            task.freed_size = space * -1
            task.install_size = 0
        else:
            task.freed_size = 0
            task.install_size = space

        for pkg_name in task.to_remove:
            if _is_critical_package(apt_cache[pkg_name]):
                print("Installer: apt - cannot remove critical package: %s" % pkg_name)
                task.info_ready_status = task.STATUS_FORBIDDEN

        if aptpkg.name in BROKEN_PACKAGES:
            print("Installer: apt- cannot execute task, package is broken: %s" % aptpkg.name)
            task.info_ready_status = task.STATUS_BROKEN

        print("For install:", task.to_install)
        print("For removal:", task.to_remove)
        print("For upgrade:", task.to_update)

        if task.info_ready_status not in (task.STATUS_FORBIDDEN, task.STATUS_BROKEN):
            task.info_ready_status = task.STATUS_OK
            task.execute = execute_transaction

    GObject.idle_add(task.info_ready_callback, task)

def sync_cache_installed_states():
    get_apt_cache(full=True)

def execute_transaction(task):
    if task.client_progress_cb != None:
        task.has_window = True

    task.transaction = MetaTransaction(task)

class MetaTransaction():
    def __init__(self, task):
        self.apt_client = aptdaemon.client.AptClient()
        self.task = task
        self.apt_transaction = None

        if task.type == "remove":
            self.apt_client.remove_packages([task.pkginfo.name],
                                            reply_handler=self._calculate_changes,
                                            error_handler=self._on_error) # dbus.DBusException
        else:
            self.apt_client.install_packages(task.to_install,
                                             reply_handler=self._calculate_changes,
                                             error_handler=self._on_error) # dbus.DBusException

    def _calculate_changes(self, apt_transaction):
        self.apt_transaction = apt_transaction
        self.apt_transaction.set_debconf_frontend("gnome")

        self.apt_transaction.simulate(reply_handler=self._confirm_changes,
                                      error_handler=self._on_error) # aptdaemon.errors.TransactionFailed, dbus.DBusException

    def _confirm_changes(self):
        try:
            if [pkgs for pkgs in self.apt_transaction.dependencies if pkgs]:
                dia = ChangesConfirmDialog(self.apt_transaction, self.task, parent=self.task.parent_window)
                res = dia.run()
                dia.hide()
                if res != Gtk.ResponseType.OK:
                    GObject.idle_add(self.task.finished_cleanup_cb, self.task)
                    return
            self._run_transaction()
        except Exception as e:
            print(e)

    def _on_error(self, error):
        if self.apt_transaction.error_code == "error-not-authorized":
            # Silently ignore auth failures

            self.task.error_message = None # Should already be none, but this is a reminder
            return
        elif not isinstance(error, aptdaemon.errors.TransactionFailed):
            # Catch internal errors of the client
            error = aptdaemon.errors.TransactionFailed(aptdaemon.enums.ERROR_UNKNOWN,
                                                       str(error))

        if self.task.progress_state != self.task.PROGRESS_STATE_FAILED:
            self.task.progress_state = self.task.PROGRESS_STATE_FAILED

            self.task.error_message = self.apt_transaction.error_details

            dia = AptErrorDialog(error)
            dia.run()
            dia.hide()
            GObject.idle_add(self.task.error_cleanup_cb, self.task)

    def _run_transaction(self):
        self.apt_transaction.connect("finished", self.on_transaction_finished)

        if self.task.has_window:
            self.apt_transaction.connect("progress-changed", self.on_transaction_progress)
            self.apt_transaction.connect("error", self.on_transaction_error)
            self.apt_transaction.run(reply_handler=lambda: None, error_handler=self._on_error)
        else:
            progress_window = AptProgressDialog(self.apt_transaction)
            progress_window.run(show_error=False, error_handler=self._on_error)

    def on_transaction_progress(self, apt_transaction, progress):
        if not apt_transaction.error:
            self.task.client_progress_cb(self.task.pkginfo, progress, False)

    def on_transaction_error(self, apt_transaction, error_code, error_details):
        if self.task.progress_state != self.task.PROGRESS_STATE_FAILED:
            self._on_error(apt_transaction.error)

    def on_transaction_finished(self, apt_transaction, exit_state):
        # finished signal is always called whether successful or not
        # Only call here if we succeeded, to prevent multiple calls
        if (exit_state == aptdaemon.enums.EXIT_SUCCESS) or apt_transaction.error_code == "error-not-authorized":
            self.task.progress_state = self.task.PROGRESS_STATE_FINISHED
            GObject.idle_add(self.task.finished_cleanup_cb, self.task)
